#define SDL_MAIN_HANDLED

#include <GL/glew.h>
#include <GL/gl.h>
#include <SDL2/SDL.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

// ---------------------------------------------------------------------

const GLdouble pi = 3.1415926535897932384626433832795;

struct vec2     { float x, y; };
struct vec3     { float x, y, z; };
struct vec4     { float x, y, z, w; };
struct TCamera  { struct vec3 pos, rot; };
struct TMouse   { int x, y, btn; };

// ---------------------------------------------------------------------

// Вершинный шейдер по умолчанию
static char defaultVertexShader[] = "#version 120\n\
attribute vec4 position;\n\
void main() {\
    gl_TexCoord[0] = gl_MultiTexCoord0;\
    gl_Position    = gl_ProjectionMatrix * gl_ModelViewMatrix * position;\
}";

// ---------------------------------------------------------------------

class App {

protected:

    SDL_Event       event;
    SDL_Window*     window;
    SDL_GLContext   glcontext;
    int             pticks = 0, width, height;
    int             frame_length = (1000 / 60);
    struct TCamera  camera;
    struct TMouse   mouse;
    int             keyb[256];
    GLuint          program_id_latest = -1;

public:

    App(int w, int h, const char* title = "OpenGL SDL2 Demo") {

        SDL_Init(SDL_INIT_EVERYTHING);

        window    = SDL_CreateWindow(title, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, w, h, SDL_WINDOW_OPENGL);
        glcontext = SDL_GL_CreateContext(window);

        // Поддержка шейдеров
        GLenum glew_status = glewInit();

        // Проверка поддержки GLEW
        if (GLEW_OK != glew_status) {

            printf("GLEW error: %s\n", glewGetErrorString(glew_status));
            if (!GLEW_VERSION_2_0) printf("No support for OpenGL 2.0 found\n");
            exit(1);
        }

        // Настройки GL
        glViewport      (0, 0, w, h);
        glEnable        (GL_DEPTH_TEST);
        glDepthFunc     (GL_LESS);          // Z-буфер, который проверяет тип LESS
        glClearDepth    (1.0f);             // Включить Z-буфер
        glBlendFunc     (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA); // Режим смешивания ALPHA
        glBlendEquation (GL_FUNC_ADD);      // Смешивание путем сложения цветов
        glEnable        (GL_TEXTURE_2D);    // Включить текстурирование
        glEnable        (GL_BLEND);         // Включить смешивание
        glEnable        (GL_CULL_FACE);     // Включить срез невидимых областей
        glCullFace      (GL_FRONT);         // Показать передние грани
        glFrontFace     (GL_CCW);           // Тип определения передной грани: против часовой стрелки

        camera.pos = {0, 0, 0};
        camera.rot = {0, 0, 0};
        mouse.btn  = 0;
        width      = w;
        height     = h;

        for (int i = 0; i < 256; i++) keyb[i] = 0;

        updateCamera();
    }

    // Обработка событий
    int poll() {

        while (true) {

            Uint32 ticks = SDL_GetTicks();

            int keypress = 0;

            while (SDL_PollEvent(& event) != 0) {

                switch (event.type) {

                    case SDL_QUIT: {
                        return 0;
                    }

                    // https://wiki.libsdl.org/SDL_MouseMotionEvent
                    case SDL_MOUSEMOTION: {

                        // Движение камеры
                        if (mouse.btn & 1) {
                            camera.rot.y += (event.motion.x - mouse.x) * 0.3;
                            camera.rot.x -= (event.motion.y - mouse.y) * 0.3;
                        }

                        mouse.x = event.motion.x;
                        mouse.y = event.motion.y;
                        break;
                    }

                    // https://wiki.libsdl.org/SDL_MouseButtonEvent
                    case SDL_MOUSEBUTTONDOWN: {

                        mouse.x    = event.motion.x;
                        mouse.y    = event.motion.y;
                        mouse.btn |= event.button.button;
                        break;
                    }

                    case SDL_MOUSEBUTTONUP: {

                        mouse.btn &= ~event.button.button;
                        break;
                    }

                    case SDL_KEYDOWN:

                        keypress = event.key.keysym.scancode;
                        if (keypress < 256) keyb[keypress] = 1;
                        break;

                    case SDL_KEYUP:

                        keypress = event.key.keysym.scancode;
                        if (keypress < 256) keyb[keypress] = 0;
                        break;
                }
            }

            // Обновление экрана
            if (ticks - pticks >= frame_length) {

                pticks = ticks;
                SDL_GL_SwapWindow(window);
                return 1;
            }

            SDL_Delay(1);
        }
    }

    // Выход из SDL
    void destroy() {

        SDL_GL_DeleteContext(glcontext);
        SDL_Quit();
    }

    // Статус нажатой клавиши
    int pressed(int key) { return keyb[key & 255]; }

    // Очистка окна от всего что там есть
    void clear(uint cl = 0x000000) {

        int r_ = (cl >> 16) & 255,
            g_ = (cl >>  8) & 255,
            b_ = (cl      ) & 255;

        glClearColor((GLfloat)r_/255., (GLfloat)g_/255., (GLfloat)b_/255., 1);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    }

    // Загрузка текста
    char* loadText(const char* filename) {

        char* data;
        FILE* fp = fopen(filename, "r");
        int   fs;

        if (!fp) {
            printf("[КРИТИЧНО] Файл '%s' не найден\n", filename);
            exit(1);
        }

        // Получение размера файла
        fseek(fp, 0, SEEK_END);
        fs = ftell(fp);
        fseek(fp, 0, SEEK_SET);

        data = (char*) malloc(fs + 1); // Выделить место (+0 zero)
        fread(data, 1, fs, fp);        // Прочесть данные
        data[ fs ] = 0;
        fclose(fp);

        return data;
    }

    // Камера
    // -----------------------------------------------------------------

    // Создание и обновление позиции камеры
    void updateCamera() {

        float fovY   = 90,
              zNear  = 0.1;

        GLdouble aspect = (GLfloat)width / (GLfloat)height;
        GLdouble fH = tan(fovY / 360 * pi) * zNear;
        GLdouble fW = fH * aspect;

        // Загрузка камеры
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        glFrustum(-fW, fW, -fH, fH, zNear, 1000);
        glDepthRange(0.1, 1000);

        // Сначала камера отодвигается, потом вращается по Z,Y,X
        glRotatef(camera.rot.x, -1, 0, 0);
        glRotatef(camera.rot.y,  0, 1, 0);
        glRotatef(camera.rot.z,  0, 0, 1);
        glTranslatef(-camera.pos.x, -camera.pos.y, camera.pos.z);

        // К матрице моделей
        glMatrixMode(GL_MODELVIEW);
    }

    // Движение камеры
    void moveCamera(vec3 v) {

        camera.pos.x += v.x;
        camera.pos.y += v.y;
        camera.pos.z += v.z;
    }

    // Рассчитать вектор, куда смотрит камера
    vec3 lookCamera(float k = 1.) {

        float x = sin(pi * camera.rot.y / 180) * k;
        float y = sin(pi * camera.rot.x / 180) * k;
        float z = cos(pi * camera.rot.y / 180) * k;

        return {x, y, z};
    }

    // Установка позиции
    void setCameraPos(vec3 p) { camera.pos = p; }
    void setCameraRot(vec3 r) { camera.rot = r; }

    // Математика
    // -----------------------------------------------------------------

    // Скалярное произведение
    float dot(vec3 a, vec3 b) {
        return a.x*b.x + a.y*b.y + a.z*b.z;
    }

    // Нормализация вектор
    vec3 normalize(vec3 a, float k = 1.) {

        float r = sqrt(dot(a, a));
        r = r ? k/r : k;
        return {r*a.x, r*a.y, r*a.z};
    }

    // Текстуры и шейдеры
    // -----------------------------------------------------------------

    // Текстура для тестов
    int stubTexture(int example = 0, int w = 256, int h = 256) {

        GLuint bgra[w*h];

        for (int i = 0; i < h; i++)
        for (int j = 0; j < w; j++) {
            bgra[i*h + j] = (i^j)*256;
        }

        return loadTexture(bgra, w, h);
    }

    // Текстура создается в памяти, назначается GLuint
    GLuint loadTexture(GLuint* data, int w, int h) {

        GLuint texture_id;

        for (uint i = 0; i < w * h; i++) {
            data[i] ^= 0xff000000;
        }

        glGenTextures   (1, (GLuint*) & texture_id);
        glBindTexture   (GL_TEXTURE_2D, texture_id);
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        glTexParameteri (GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
        glTexImage2D    (GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

        return texture_id;
    }

    // Загрузка шейдера с диска
    GLuint loadShader(const char* fragment, const char* vertex = NULL) {

        int     link_ok, len;
        char    log_buffer[1024];
        GLuint  v_shader, f_shader;

        // Загрузка шейдеров в память
        char*   v_shader_src;
        char*   f_shader_src = loadText(fragment);

        // Создать вершинный и фрагментные шейдеры
        v_shader = glCreateShader(GL_VERTEX_SHADER);
        f_shader = glCreateShader(GL_FRAGMENT_SHADER);

        // Передача исходного кода если NULL
        if (vertex == NULL) {

            int lens = strlen(defaultVertexShader);
            v_shader_src = (char*) malloc(lens + 1);
            strcpy(v_shader_src, defaultVertexShader);
            glShaderSource(v_shader, 1, (const char**) & v_shader_src, NULL);
            free(v_shader_src);
        }
        // Либо шейдер задан
        else {

            v_shader_src = loadText(vertex);
            glShaderSource(v_shader, 1, (const char**) & v_shader_src, NULL);
        }

        glShaderSource(f_shader, 1, (const char**) & f_shader_src, NULL);

        // Компиляция шейдера
        glCompileShader(f_shader);
        glCompileShader(v_shader);

        // Получить ID программы
        GLuint program = glCreateProgram();

        // Прикрепить шейдеры к программе
        glAttachShader(program, v_shader);
        glAttachShader(program, f_shader);

        // Линкуем шейдерную программу
        glLinkProgram( program );

        // Проверка статуса линковки программы
        glGetProgramiv(program, GL_LINK_STATUS, &link_ok);

        if (!link_ok) {

            printf("[КРИТИЧНО] Программа не слинкована\n");

            glGetShaderInfoLog(v_shader, sizeof(log_buffer), &len, log_buffer);
            printf("  [VERTEX] %s\n", log_buffer);

            glGetShaderInfoLog(f_shader, sizeof(log_buffer), &len, log_buffer);
            printf("  [FRAGMENT] %s\n", log_buffer);

            glGetProgramInfoLog(program, sizeof(log_buffer), &len, log_buffer);
            printf("  [PROGRAM] %s\n", log_buffer);

            exit(1);
        }

        // Не освобождать память, если не занято было
        if (vertex != NULL) free(v_shader_src);
        free(f_shader_src);

        // Сразу же использовать программу после загрузки
        glUseProgram(program);

        // Запись последней программы
        program_id_latest = program;

        return program;
    }

    // Задать униформу
    GLint uniform1f(GLuint _program_id, const char* name, float value) {

        GLint loc = glGetUniformLocation(_program_id, name);
        if (loc != -1) {
            glUniform1f(loc, value);
        }

        return loc;
    }
};
